import abc

from gradema.section import Section
from gradema.test import TestReporter, TestResult


class GraderReporter(abc.ABC):
    """
    A GraderReporter provides functionality to report on things occurring during the grading process.
    """

    @property
    @abc.abstractmethod
    def test_reporter(self) -> TestReporter:
        pass

    @abc.abstractmethod
    def report_start(self, max_points: int) -> None:
        pass

    @abc.abstractmethod
    def report_test_result(self, result: TestResult, points: int, max_points: int) -> None:
        pass

    @abc.abstractmethod
    def subsection(self, section: Section) -> "GraderReporter":
        """
        Creates a similar GraderReporter containing the passed subsection and updated attributes to correctly report on a subsection.

        :param section: A subsection of the current section being reported on.
        :return: The modified GraderReporter
        """
        pass



import sys

from rich.console import Console
from rich.text import Text

from gradema.grader import GraderReporter
from gradema.grader.console import ConsoleGraderReporter
from gradema.section import Section
from gradema.test import Test, FractionalTestResult, PercentTestResult


def point_sections(max_points: int, sections: list[Section]) -> list[int]:
    point_values_sum = 0
    point_values: list[int] = [0 for _ in range(len(sections))]
    weight_total = 0.0
    evenly_weighted_count = 0
    for i, section in enumerate(sections):
        if section.is_pointed or section.is_ungraded:
            assert section.points is not None
            point_values[i] = section.points
            point_values_sum += section.points
        elif section.is_weighted:
            assert section.weight is not None
            weight_total += section.weight
        else:
            assert section.is_evenly_weighted
            evenly_weighted_count += 1
    if evenly_weighted_count > 0:
        evenly_weighted_weight = (1 - weight_total) / evenly_weighted_count
        if evenly_weighted_weight <= 0:
            raise ValueError(
                f"weight_total: {weight_total} which is invalid when you have evenly weighted section. "
                f"Make sure it is less than 1 so the weight can be distributed evenly among those sections."
            )
        weight_divisor = 1.0
    else:
        evenly_weighted_weight = 0.0
        weight_divisor = weight_total

    remaining_points = max_points - point_values_sum
    if remaining_points < 0:
        raise ValueError(f"The sum of all the pointed sections are higher than max points. max_points: {max_points} sum: {sum(point_values)}")

    points_assigned = 0  # the sum of the points assigned to weighted or evenly weighted sections
    running_weight_sum = 0.0
    for i, section in enumerate(sections):
        if section.is_weighted or section.is_evenly_weighted:
            if section.is_weighted:
                assert section.weight is not None
                weight = section.weight / weight_divisor
            else:
                weight = evenly_weighted_weight
            running_weight_sum += weight
            target_points_assigned = remaining_points * running_weight_sum
            points = int(target_points_assigned - points_assigned)
            point_values[i] = points
            points_assigned += points

    assert points_assigned == remaining_points
    return point_values


def welcome(console: Console) -> None:
    console.print("Welcome to Gradema!")
    console.print("This script will grandma you through completing this programming assignment")
    console.line()
    console.print(f"    [magenta]Magenta[/magenta] is bash/shell commands.")
    console.print(f"    [green]Green[/green] is used to indicate a pass.")
    console.print(f"    [red]Red[/red] is used to indicate a fail.")
    console.print(f"    [orange_red1]Orange[/orange_red1] marks the text of a header for a block of tests of a single type.")
    console.print(f"    [cyan]Cyan[/cyan] draws your attention to something you have to do..")
    console.line()
    console.print(Text.assemble("Actually ", ("read ALL the output generated by this script!", "cyan")))

    console.print(f"[blue]{'=' * console.width}[/blue]")


def grade_test(reporter: GraderReporter, max_points: int, test: Test) -> int:
    reporter.report_start(max_points)
    result = test.run(reporter.test_reporter)
    if isinstance(result, FractionalTestResult):
        points = max_points * result.success_count // result.total_count
    else:
        assert isinstance(result, PercentTestResult)
        points = int(result.percent * max_points)

    reporter.report_test_result(result, points, max_points)
    return points


def grade_section(reporter: GraderReporter, max_points: int, section: Section) -> int:
    if isinstance(section.node, Test):
        return grade_test(reporter, max_points, section.node)
    assert isinstance(section.node, list)
    sections: list[Section] = section.node
    assert len(sections) > 0
    assert isinstance(sections[0], Section)
    point_values = point_sections(max_points, sections)

    reporter.report_start(max_points)
    total_points_earned = 0
    for i, section in enumerate(sections):
        points = point_values[i]
        subsection_reporter = reporter.subsection(section)
        points_received = grade_section(subsection_reporter, points, section)
        total_points_earned += points_received
    return total_points_earned


def run_grader(args: list[str], section: Section) -> int:
    if not section.is_pointed:
        raise ValueError("The root section must be pointed!")
    assert section.points is not None
    # Useful documentation here: https://rich.readthedocs.io/en/latest/markup.html
    console = Console(soft_wrap=True)
    welcome(console)

    interactive = True
    debug_mode = True
    try:
        console.print("Press [cyan]Enter[/cyan] now to continue")
        input()
    except EOFError:
        interactive = False
        debug_mode = False

    if interactive:
        console.print(
            """Which of the following modes do you want to run in?
    1) debug+grade mode (d)
    2) grade only mode (g)
Type 'd' or 'g', then hit [cyan]Enter[/cyan]."""
        )
        mode_input = input()
        debug_mode = mode_input == "" or mode_input[0].lower() != "g"

    reporter = ConsoleGraderReporter(
        section,
        console,
        sys.stdout,
        sys.stderr,
        interactive=interactive,
        debug_mode=debug_mode,
        depth=0,
    )
    points = grade_section(reporter, section.points, section)
    passing = points >= 70
    console.line()
    console.print("=" * console.width, style="green" if passing else "red")
    console.print(f"Your total grade is:\n{points}")
    console.line()
    if points == 100:
        console.print("It's perfect! Congratulations!", style="green")
        return 0
    elif passing:
        console.print("You're passing; decide how much you want the extra points!", style="green")
        return 0
    else:
        console.print("You're not passing yet.", style="red")
        return 1



import importlib
from pathlib import Path

import editorconfig  # type: ignore

from gradema.test import Test
from gradema.test._command import CommandTest
from gradema.test._stdio import StdioTest


def create_python_test(module_name: str) -> Test:
    try:
        # TODO importing the module could run student code in the autograder's process. Maybe we shouldn't do this
        importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise ValueError(f"Module {module_name} not found")

    return CommandTest(["python3", "-m", module_name], ["pudb", "-m", module_name])


def create_python_stdio_test(module_name: str, test_identifier: str, input_file: Path, goal_file: Path) -> Test:
    """
    Creates a traditional stdio test

    :param module_name:
    :param test_identifier:
    :param input_file:
    :param goal_file:
    :return:
    """
    command = ["python3", "-m", module_name]
    return StdioTest(command, test_identifier, input_file, goal_file)


def create_python_arg_test(module_name: str, test_identifier: str, args: list[str], goal_file: Path) -> Test:
    command = ["python3", "-m", module_name] + args
    return StdioTest(command, test_identifier, None, goal_file)


def create_python_format_check_from_path(path: str) -> Test:
    line_length = editorconfig.get_properties(str((Path(path) / "some_random_file.py").absolute())).get("max_line_length")
    command = ["black", "--check"]
    if line_length is not None:
        command.append(f"--line-length={line_length}")
    command.append(path)
    return CommandTest(command)


def create_python_format_check() -> Test:
    return create_python_format_check_from_path(".")


def create_python_type_check() -> Test:
    return CommandTest(["mypy", "--strict", "--disallow-any-explicit", "."])



import copy
import dataclasses
import difflib
import io
import shlex
import shutil
import subprocess
import traceback
from pathlib import Path
from typing import TextIO, Optional

from rich import markup
from rich.console import Console
from rich.prompt import Prompt

from gradema.grader import GraderReporter
from gradema.grader.console._util import indentation, arrow, header, indent_text
from gradema.section import Section
from gradema.test import (
    TestReporter,
    CommandData,
    Test,
    TestResult,
    FractionalTestResult,
    PercentTestResult,
    StdioCommandData,
)


BACKSLASH = markup.escape("\\")
TEST_DIRECTORY = Path("build/test")


def has_vim_installed() -> bool:
    return shutil.which("vim") is not None


def vim_diff(goal_file: Path, output_file: Path, html_output: Path) -> None:
    """
    Creates a diff using vim.
    Note that calling this function takes a while to complete.

    :param goal_file:
    :param output_file:
    :param html_output:
    """
    subprocess.run(
        [
            "vim",
            "-d",
            str(goal_file),
            str(output_file),
            "-c",
            "highlight DiffAdd ctermbg=Grey ctermfg=White",
            "-c",
            "highlight DiffDelete ctermbg=Grey ctermfg=Black",
            "-c",
            "highlight DiffChange ctermbg=Grey ctermfg=Black",
            "-c",
            "highlight DiffText ctermbg=DarkGrey ctermfg=White",
            "-c",
            "TOhtml",
            "-c",
            f"w! {str(html_output)}",
            "-c",
            "qa!",
        ],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def html_diff(left_path: Path, right_path: Path, html_output_path: Path):
    with left_path.open('r') as left, right_path.open('r') as right:
        diff = difflib.HtmlDiff().make_file(
            left.readlines(),
            right.readlines(),
            str(left_path),
            str(right_path),
        )
        with html_output_path.open('w') as f_out:
            f_out.write(diff)


@dataclasses.dataclass
class ConsoleTestReporter(TestReporter):
    reporter: "ConsoleGraderReporter"

    def log(self, message: str) -> None:
        self.reporter.console.print(markup.escape(message))

    def log_unexpected_exception(self, exception: BaseException) -> None:
        indent = indentation(self.reporter.depth)
        buffer = io.StringIO()
        traceback.print_exception(exception, file=buffer)
        indented_text = indent_text(self.reporter.depth, buffer.getvalue())
        # make end="" because the exception text should have a trailing newline
        self.reporter.console.print(f"{indent}Unexpected exception:\n{indented_text}", highlight=False, end="")

    def report_command(self, command: list[str]) -> CommandData:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(f"{indent}Running command: [magenta]{markup.escape(shlex.join(command))}[/magenta]")
        return CommandData(self.reporter.stdout, self.reporter.stderr)

    def report_stdio_command(self, command: list[str], test_identifier: str, input_file: Optional[Path]) -> StdioCommandData:
        indent = indentation(self.reporter.depth)
        indent_plus = indentation(self.reporter.depth + 1)
        directory = TEST_DIRECTORY / "stdio/outputs"
        directory.mkdir(parents=True, exist_ok=True)
        output_file = directory / f"{test_identifier}.txt"
        self.reporter.console.print(f"{indent}Command to be run: [magenta]{markup.escape(shlex.join(command))}[/magenta]")
        self.reporter.console.print(f"{indent}Command to be run including I/O redirection:")
        self.reporter.console.print(
            f"{indent}[magenta]{markup.escape(shlex.join(command))} {BACKSLASH}[/magenta]"
            + (f"\n{indent_plus}[magenta]< {markup.escape(shlex.quote(str(input_file)))} {BACKSLASH}[/magenta]" if input_file is not None else "")
            + f"\n{indent_plus}[magenta]> {markup.escape(shlex.quote(str(output_file)))}[/magenta]"
        )

        return StdioCommandData(output_file, self.reporter.stderr)

    def maybe_launch_debugger(self, command: list[str]) -> None:
        indent = indentation(self.reporter.depth)
        if self.reporter.debug_mode:
            self.reporter.console.print(f"{indent}This test did not complete successfully. We are going to launch the debugger now.")
            self.reporter.console.print(f"{indent}Debug command: [magenta]{markup.escape(shlex.join(command))}[/magenta]")
            launch_debugger = True
            while True:
                response = Prompt.ask(
                    prompt=f"{indent}Press [cyan]Enter[/cyan] to launch the debugger, or type 's' and then [cyan]Enter[/cyan] to skip",
                    console=self.reporter.console,
                )
                if response == "":
                    break
                if response[0].lower() == "s":
                    launch_debugger = False
                    break
            if launch_debugger:
                subprocess.run(command)

    def report_diff_result(self, test_identifier: str, goal_file: Path, output_file: Path, similarity: float, fuzzy: bool) -> None:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(f"{indent}Diff between [yellow]{markup.escape(str(goal_file))}[/yellow] and [yellow]{markup.escape(str(output_file))}[/yellow]")
        # TODO Customize stdio directory here
        directory = TEST_DIRECTORY / "stdio/diffs"
        directory.mkdir(parents=True, exist_ok=True)
        html_output = directory / f"{test_identifier}.html"
        html_diff(goal_file, output_file, html_output)
        self.reporter.console.print(f"{indent}Diff available to view at file://{html_output.resolve().absolute()}")

    def report_file_exists_test(self, file: Path, expected: str) -> None:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(f"{indent}Checking for the existence of '{file}' and its containing type of data: {expected}")

    def report_file_exists_test_result(self, file: Path, expected: str, is_correct: bool, actual: Optional[str]) -> None:
        indent_plus = indentation(self.reporter.depth + 1)
        if is_correct:
            self.reporter.console.print(f"{indent_plus}[green]Actual type matches expected type![/green] Actual: {actual}")
        elif actual is not None:
            self.reporter.console.print(f"{indent_plus}[red]Actual type DOES NOT match expected type![/red] Actual: {actual}")
        else:
            self.reporter.console.print(f"{indent_plus}[red]File does not exist![/red]")


@dataclasses.dataclass
class ConsoleGraderReporter(GraderReporter):
    section: Section
    console: Console
    stdout: TextIO
    stderr: TextIO
    interactive: bool
    debug_mode: bool
    depth: int

    @property
    def test_reporter(self) -> TestReporter:
        return ConsoleTestReporter(self)

    def __report_test(self, max_points: int) -> None:
        test = self.section.node
        assert isinstance(test, Test)
        indent = indentation(self.depth)
        self.console.line()
        if self.depth == 0:
            self.console.print(
                (indent + header(self.depth, self.console.width))[: self.console.width],
                style="blue",
            )
        if max_points == 0:
            self.console.print(f"{arrow(self.depth)}[yellow]{markup.escape(self.section.name)}[/yellow] (ungraded)")
        else:
            self.console.print(f"{arrow(self.depth)}[yellow]{markup.escape(self.section.name)}[/yellow] is worth {max_points} points")
        self.console.print(
            (indent + header(self.depth, self.console.width))[: self.console.width],
            style="blue",
        )

    def __report_section(self, max_points: int) -> None:
        assert isinstance(self.section.node, list)
        sections: list[Section] = self.section.node
        assert len(sections) > 0
        assert isinstance(sections[0], Section)

        indent = indentation(self.depth)
        self.console.line()
        if self.depth == 0:
            self.console.print(
                (indent + header(self.depth, self.console.width))[: self.console.width],
                style="blue",
            )
        self.console.print(f"{arrow(self.depth)}[orange_red1]{self.section.name}[/orange_red1] is worth {max_points} points and has {len(sections)} parts")
        self.console.print(
            (indent + header(self.depth, self.console.width))[: self.console.width],
            style="blue",
        )

    def report_start(self, max_points: int) -> None:
        if isinstance(self.section.node, Test):
            self.__report_test(max_points)
        else:
            self.__report_section(max_points)

    def report_test_result(self, result: TestResult, points: int, max_points: int) -> None:
        indent = indentation(self.depth)
        if max_points == 0:
            self.console.print(f"{indent}Ungraded step completed successfully.")
        elif isinstance(result, FractionalTestResult):
            if result.total_count == 1:
                if result.success_count == 0:
                    self.console.print(
                        f"{indent}FAIL gives you {points} / {max_points} points",
                        style="red",
                    )
                else:
                    self.console.print(
                        f"{indent}SUCCESS gives you {points} / {max_points} points",
                        style="green",
                    )
            else:
                self.console.print(
                    f"{indent}Result is {result.success_count} / {result.total_count} which gives you {points} / {max_points} points",
                    style="green" if points > max_points // 2 else "red",
                )
        else:
            assert isinstance(result, PercentTestResult)
            self.console.print(f"Result is {result.percent:.2%} which gives you {points}")

    def subsection(self, section: Section) -> "GraderReporter":
        r = copy.copy(self)
        r.section = section
        r.depth += 1
        return r



import subprocess

from gradema.test import Test, TestReporter, TestResult, FractionalTestResult


class CommandTest(Test):
    def __init__(self, command: list[str], debug_command: list[str] | None = None):
        self.command = command
        self.debug_command = debug_command

    def run(self, reporter: TestReporter) -> TestResult:
        data = reporter.report_command(self.command)

        completed = subprocess.run(
            self.command,
            stdin=subprocess.DEVNULL,
            stdout=data.stdout,
            stderr=data.stderr,
        )
        # TODO prevent students from doing sys.exit(0)
        if completed.returncode == 0:
            return FractionalTestResult(1, 1)
        if self.debug_command is not None:
            reporter.maybe_launch_debugger(self.debug_command)
        return FractionalTestResult(0, 1)



def indentation(depth: int) -> str:
    return "  " * depth


def indent_text(depth: int, text: str) -> str:
    needs_indent = True
    indent = indentation(depth)
    r = ""
    for c in text:
        if c == "\n":
            needs_indent = True
        elif c != "\r" and needs_indent:
            r += indent
            needs_indent = False
        r += c
    return r


def arrow(depth: int) -> str:
    return "--" * (depth - 1) + "> " * (1 if depth else 0)


def header(depth: int, width: int) -> str:
    if depth == 0 or depth == 1:
        return "=" * width
    elif depth == 1:
        return "-" * width
    elif depth == 2:
        return "-" * width
    elif depth == 3:
        return '"' * width
    else:
        return "~" * width



import filecmp
import subprocess
from pathlib import Path
from typing import Optional, TextIO

from gradema.test import Test, TestReporter, TestResult, FractionalTestResult


class StdioTest(Test):
    """
    A Standard Input/Output Test

    Traditionally, a stdio test would have no arguments passed to the command, it would have an input file, and a goal file.
    Output is piped to some output file that is not committed to version control.

    This new stdio test is very similar, but it can also do what a traditional arg test could do:
    take no input, and have custom arguments passed to the command.
    If you don't want anything directed into the command's stdin, then just assign None to input_file.
    If you want custom arguments passed to the command, give those custom arguments to the command argument (you may have to do that manually).
    """

    def __init__(self, command: list[str], test_identifier: str, input_file: Optional[Path], goal_file: Path):
        self.command = command
        self.test_identifier = test_identifier
        self.input_file = input_file
        self.goal_file = goal_file

    def run(self, reporter: TestReporter) -> TestResult:
        data = reporter.report_stdio_command(self.command, self.test_identifier, self.input_file)
        output_file = data.stdout_file
        stdin: Optional[TextIO] = None
        try:
            if self.input_file is not None:
                stdin = self.input_file.open("r")
            with output_file.open("w") as stdout:
                completed = subprocess.run(
                    self.command,
                    stdin=stdin if stdin is not None else subprocess.DEVNULL,
                    stdout=stdout,
                    stderr=data.stderr,
                )
        finally:
            if stdin is not None:
                stdin.close()
        if completed.returncode != 0:
            reporter.log(f"Process had non-zero exit code: {completed.returncode}. Not going to execute a diff until the program can complete successfully.")
            return FractionalTestResult(0, 1)

        exactly_the_same = filecmp.cmp(self.goal_file, output_file)
        # TODO fuzzy diff stuff here
        reporter.report_diff_result(self.test_identifier, self.goal_file, output_file, 1.0 if exactly_the_same else 0.0, fuzzy=False)
        if exactly_the_same:
            return FractionalTestResult(1, 1)

        return FractionalTestResult(0, 1)



from gradema.test import Test, TestReporter, TestResult, FractionalTestResult


class DummyTest(Test):

    @property
    def name(self) -> str:
        return "Dummy Test"

    def run(self, reporter: TestReporter) -> TestResult:
        return FractionalTestResult(1, 1)


DUMMY_TEST = DummyTest()


def dummy_test() -> Test:
    return DUMMY_TEST
